import { TextureDimension, TextureFormat } from '../enum.js'
import { WebGPUObjectError } from '../utils/errors.js'
import { ResourceType } from '../utils/internal-enums.js'
import { textureToImageDimension } from '../utils/wgsl-to-wgpu.js'

/** @import { BindGroupEntry } from '../core/graphics-device.js' */
interface UploadTextureInfo {
  destination: GPUTexelCopyTextureInfo
  dataLayout: GPUTexelCopyBufferLayout
  size: GPUExtent3DStrict
}

export class Texture {
  static _defaultUsage = GPUTextureUsage.TEXTURE_BINDING
    | GPUTextureUsage.COPY_DST
    | GPUTextureUsage.RENDER_ATTACHMENT

  _device: GPUDevice
  _handle: GPUTexture

  _defaultView: GPUTextureView | undefined

  get handle() {
    return this._handle
  }

  get width() {
    return this._handle.width
  }

  get height() {
    return this._handle.height
  }

  get format() {
    return this._handle.format
  }

  get depthOrArrayLayers() {
    return this._handle.depthOrArrayLayers
  }

  get usage() {
    return this._handle.usage
  }

  get dimension() {
    return this._handle.dimension
  }

  get mipLevelCount() {
    return this._handle.mipLevelCount
  }

  get label() {
    return this._handle.label
  }

  get resourceType() {
    return ResourceType.TextureView
  }

  constructor(device: GPUDevice, texture: GPUTexture) {
    this._device = device
    this._handle = texture
  }

  static create(device: GPUDevice, descriptor: GPUTextureDescriptor) {
    try {
      return new Texture(
        device,
        device.createTexture(descriptor)
      )
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  static _generateMipLevels(size: number[]) {
    const max = Math.max.apply(undefined, size)
    return 1 + Math.log2(max) | 0
  }


  static async fromUrl(device: GPUDevice, url: string | URL, descriptor: GPUTextureDescriptor) {
    try {
      const response = await fetch(url)

      if (!response.ok) {
        throw new Error(`Failed to fetch remote resource: ${response.statusText}`)
      }

      const usage = descriptor.usage || Texture._defaultUsage
      const dimension = descriptor.dimension ? textureToImageDimension(descriptor.dimension) : TextureDimension['2d']

      const blob = await response.blob()
      const bitmap = await createImageBitmap(blob)
      const size = [bitmap.width, bitmap.height]

      const desc = {
        usage,
        dimension,
        size,
        format: descriptor.format || TextureFormat.Rgba8unorm,
        mipLevelCount: descriptor.mipLevelCount || Texture._generateMipLevels(size),
        ...descriptor,
      }

      const texture = Texture.create(device, desc)

      texture.upload(bitmap)

      return texture
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  static createRenderTarget(device: GPUDevice, size: GPUExtent3DStrict, format: GPUTextureFormat, descriptor: GPUTextureDescriptor) {
    const usage = descriptor.usage || Texture._defaultUsage

    return Texture.create(device, {
      size,
      format,
      usage,
      ...descriptor
    })
  }


  upload(source: GPUAllowSharedBufferSource, options?: UploadTextureInfo) {
    const mipLevel = options.destination.mipLevel || 0
    const size = options.size || [
      Math.max(1, this.width >> mipLevel),
      Math.max(1, this.height >> mipLevel),
      this.depthOrArrayLayers
    ]

    try {
      this._device.queue.writeTexture(
        { ...options.destination, texture: this._handle },
        source,
        options.dataLayout,
        size
      )
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  generateMipmaps(_commandEncoder: GPUCommandEncoder) {
    // TODO: use MipGenerator
  }

  createDefaultView(descriptor?: GPUTextureViewDescriptor) {
    if (!descriptor && this._defaultView) {
      return this._defaultView
    }

    try {
      const view = this._handle.createView(descriptor)

      if (!descriptor) {
        this._defaultView = view
      }

      return view
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  getView() {
    return this.createDefaultView()
  }

  toBindingResource() {
    return this.getView()
  }

  destroy() {
    this._handle?.destroy()
    this._handle = undefined
    this._defaultView = undefined
  }
}

